/**
 *
 * Mozilla Public License Notice
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1  (the "License"); you may  not use this  file except in
 * compliance with the  License. You may obtain a  copy of the License
 * at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the  License  for  the   specific  language  governing  rights  and
 * limitations under the License.
 *
 * The  Original Code and  Idea is  the Hit-a-Hint  Mozilla extension.
 * The Initial  Developer of the Original  Code and the  Idea is Pekka
 * P. Sillanpaa.  Portions created  by Initial Developer are Copyright
 * (C) 2004.  All Rights Reserved.
 *
 * Contributor(s): Pekka Sillanpaa, Paul Stone
 *
 */

var hintElemArray = null; // array of hint elements
var hintableElemArray = null;
var hahElemArray = null;
var linkNumString = ""; // the typed link number is in this string
var hahQuickDisabled = false;
var isHahModeEnabled = false; // is typing mode on
var regenElemList = false;
var isFirstTime = true;
var areCoordsValid = true;
var hahDocCount = 0; // counter for frames in the document
var hahFrameArray = null;
var hahFAYTState = null;

var isMagicKeyPressed = false;

// not (yet?) in preferences
var BACKSPACE_KEY = 8; // backspace
var FINISH_KEY = 13; // enter
var STOP_KEY = 27; // esc

var HINT_PREFIX = 'hint'; // prefix for the hint id

var loads = 0;

// for BECL (background element coordidate loader)
var hahBECLwin = 0;
var hahBECLtag = 0;
var hahBECLelem = 0;

var getHintableElemsRunning = false;

var oWindowDocument = null;

var BECLid = null;

////////////////////////////////////////////////////////////////////////////////
// logging
////////////////////////////////////////////////////////////////////////////////

var gConsoleService = Components.classes['@mozilla.org/consoleservice;1']
                    .getService(Components.interfaces.nsIConsoleService);

function hahLogMessage(aMessage)
{
  gConsoleService.logStringMessage('hah: ' + aMessage);
}

////////////////////////////////////////////////////////////////////////////////
// configuration and initialization related functions
////////////////////////////////////////////////////////////////////////////////

function hahLoad()
{
  if (isFirstTime) {
    hahPrefs.setDefaultPreferences();
    isFirstTime = false;
  }
  if(!hahPrefs.get(HAH_PREFS.ENABLED) || !hahPrefs.get(HAH_PREFS.STATUSICON_ENABLED)) {
    hahStatusIcon('', false);
  } else {
    hahStatusIcon("enabled", true);
  } 
  
  if(!hahQuickDisabled) {
    window.addEventListener("keydown", hahOnKeydown, true);
    window.addEventListener("keyup", hahOnKeyup, true);
    window.addEventListener("keypress", hahOnKeypress, true);
  }

  if (isHahModeEnabled) hahRemoveHints();

  isHahModeEnabled = false;

  areCoordsValid = false;

  window.setTimeout("getHintableElems()", 10);
  oWindowDocument = window.document;
}

function hahBlurFocused() {
  var elt = document.commandDispatcher.focusedElement;
  if (elt != null) elt.blur();
  //  getDocument().body.focus();
}

function hahResize() 
{
  areCoordsValid = false;
  if (isHahModeEnabled) {
    hahRemoveHints();
    hahShowHints();
  }
}

function hahQuickToggle() 
{
  if (hahQuickDisabled) {
    hahQuickDisabled = false;
    hahLoad();
    hahStatusIcon("enabled", true);
  }
  else {
    hahQuickDisabled = true;
    
    window.removeEventListener("keydown", hahOnKeydown, true);
    window.removeEventListener("keyup", hahOnKeyup, true);
    window.removeEventListener("keypress", hahOnKeypress, true);
  
    hahStatusIcon("disabled", true);
  }
}

////////////////////////////////////////////////////////////////////////////////
// hint activating and loading related functions
////////////////////////////////////////////////////////////////////////////////

function hahShowHints()
{
  generateHints();

  hahStatusIcon("enabled", true);
}

function hahRemoveHints()
{
  for (var i = 0; i < hintElemArray.length; i++) {
    var hint = hintElemArray[i];
    hint.ownerDocument.body.removeChild(hint);
    
    //hintElemArray[i].parentNode.removeChild(hint);
  }
}


/**
 * generateHints()
 *
 * iterates the hintableElemArray (generated in getHintableItems)
 * and create a hint element for each visible hintable element
 * 
 * hintableElemArray is a three-dimensional array, where first
 * array is for frames, second for tags, and third for elements
 *
 */

function generateHints()
{ 
  var linkCount = 0;

  var doc, styleString, hintElemSpan;

  var area = new Array(4);

  hintElemArray = new Array();
  hahElemArray = new Array();

  var win = window._content;

  styleString = hahPrefs.get(HAH_PREFS.HINT_STYLE);  

  for (j = 0; j < hintableElemArray.length; j++) {
    // move the area  a little bit to up-left so that also links
    // that are just a bit hidden are shown
    area[0] = win.pageXOffset - 5;
    area[1] = win.pageYOffset - 5;
    area[2] = area[0] + win.innerWidth;
    area[3] = area[1] + win.innerHeight;
    
    doc = win.document;
    
    hintElemSpan = doc.createElement('SPAN');
    hintElemSpan.setAttribute('style', styleString);
    hintElemSpan.style.position = 'absolute';

    for (k = 0; k < hintableElemArray[j].length; k++) {
      for (i = 0; i < hintableElemArray[j][k].length; i++) {

	elem = hintableElemArray[j][k][i];
	
	if (!areCoordsValid) {
	  getElemCoords(elem);
	}
	
	if (elem.absoY > area[3] || elem.absoY < area[1] || elem.absoX > area[2] || elem.absoX < area[0])
	  continue;
	
	if (elem.offsetWidth == 0 && elem.offsetHeight == 0)
	  continue;


	cs = doc.defaultView.getComputedStyle(elem, '');
	
	if (cs.visibility == "hidden")
	  continue;
	
	var hintElem = hintElemSpan.cloneNode(false);
	hintText = formatHint(linkCount)
	hintElem.appendChild(doc.createTextNode(hintText));
	hintElem.style.left = elem.absoX + 'px';
	hintElem.style.top = elem.absoY + 'px';
	doc.body.appendChild(hintElem);
	
	hahElemArray[linkCount] = elem;
	hintElemArray[linkCount] = hintElem; // save the hintelem so that it can be referenced
	linkCount++;
      }
    }
 
    win = hahFrameArray[j];
  }
  
  areCoordsValid = true;
}
/**
 * getHintableItems
 *
 * gets all elements listed in the tag list in the preferences
 * this is done for every frame recursively as well
 *
 */
function getHintableElems()
{
  getHintableElemsRunning = true;

  var linkCount = 0;
  var hintableElems = null;


  var elem = null, cs = null, hintElem = null;

  var tags = hahPrefs.get(HAH_PREFS.HINT_TAGS).split(';');
  var tag;  

  if (tags.length == 0) {
    return;
  }

  var hahDocCount = 0;

  var win = window._content;

  hintableElemArray = new Array();
  
  hahFrameArray = new Array();
  for (i = 0; i < win.frames.length; i++)
    hahFrameArray.push(win.frames[i]);

  while (win) {
    var doc = win.document;
    
    var hintableTagArray = new Array(); 
    
    for (tag = 0; tag < tags.length; tag++) {
      hintableElems = doc.getElementsByTagName(tags[tag]);      
      if (hintableElems.length > 0) {
	hintableTagArray.push(hintableElems);  
      }
    }

    hintableElemArray.push(hintableTagArray);
    
    win = hahFrameArray[hahDocCount++];
    
    if (win)
      for (i = 0; i < win.frames.length; i++)
	hahFrameArray.push(win.frames[i]);
  }

  startBackgroundElemCoordLoader();

  getHintableElemsRunning = false;
}

/**
 * Starts the coordinateloader in background, because this reduce
 * the time of hints appearing first time.
 * Getting the elements' absolute coordinates is an expensive job,
 * and in pages with thousands of links this feature adds convenience
'*/
function startBackgroundElemCoordLoader() {
  hahBECLwin = 0;
  hahBECLtag = 0;
  hahBECLelem = 0;

  if (BECLid != null) {
    window.clearTimeout(BECLid);
  }
  BECLid = window.setTimeout("backgroundElemCoordLoader()", 1000);
}

/**
 * This function calls recursively itself in the background until
 * all the elements' coordinates have been loaded
 *
 */
function backgroundElemCoordLoader() // "BECL"
{
  //  hahLogMessage(hahBECLwin + ","+hahBECLtag+","+hahBECLelem);

  if (hintableElemArray[hahBECLwin].length == 0) {
    hahBECLtag = 0;
    hahBECLelem = 0; 
    hahBECLwin++;
  } else {
    var elem = hintableElemArray[hahBECLwin][hahBECLtag][hahBECLelem];
    getElemCoords(elem);
    
    if (hahBECLelem < hintableElemArray[hahBECLwin][hahBECLtag].length - 1) {
      hahBECLelem++;
    } else if (hahBECLtag < hintableElemArray[hahBECLwin].length - 1) {
      hahBECLelem = 0;
      hahBECLtag++;
    } else if (hahBECLwin < hintableElemArray.length - 1) {
      hahBECLtag = 0;
      hahBECLelem = 0;
      hahBECLwin++;
    } else {
      BECLid = null;
      areCoordsValid = true;
      //      hahLogMessage("Coordinates pre-loaded");
      return;
    }
  }
  if (areCoordsValid || isHahModeEnabled) {
    BECLid = null;    
  } else {
    window.setTimeout("backgroundElemCoordLoader()", 1);  
  }
}

function formatHint(hintNum) {
  var numbers = hahPrefs.get(HAH_PREFS.NUMBER_KEYS);
  var str = hintNum.toString(numbers.length); // turn hintNum into a base(length) number
  
  if (hahPrefs.get(HAH_PREFS.HINT_FORCE_NUMERIC)) {
    return str;
  } else {
    // map the number onto the chars in the numbers string
    var result = '';
    for (var i=0; i<str.length; i++)
      result += (numbers.charAt(parseInt(str[i], numbers.length))).toUpperCase();;
    return result;
  }
}

function getElemCoords(elem)
{
  if (elem.absoX == undefined || !areCoordsValid) {
    var elemTemp = elem;
    elem.absoX = 0;
    elem.absoY = 0;
    while (elemTemp.tagName != 'BODY') {
      elem.absoX += elemTemp.offsetLeft;
      elem.absoY += elemTemp.offsetTop;
      elemTemp = elemTemp.offsetParent;

      if (elemTemp == null)
	break;

      if (elemTemp.absoX != undefined) {
	elem.absoX += elemTemp.absoX;
	elem.absoY += elemTemp.absoY;
	break;
      }
    }
  }

  if (elem.tagName == 'A') {
    var t = elem.firstChild; // we use image coordinates instead of the anchor
    if (t != null && t.tagName != undefined && t.tagName == 'IMG') {
      elem.absoX += t.offsetLeft - elem.offsetLeft;
      elem.absoY += t.offsetTop - elem.offsetTop;
    } 
  } else if (elem.tagName == 'AREA') {
    // find imagemap's coordinates
    var coords = elem.getAttribute("coords").split(",");
    elem.absoX += Number(coords[0]);
    elem.absoY += Number(coords[1]);
  }
}

function hahChangeHintFocus(linkNumString, oldLinkNumString)
{
  var nums = hahPrefs.get(HAH_PREFS.NUMBER_KEYS);
  var hintId = parseInt(linkNumString, nums.length);
  var oldHintId = parseInt(oldLinkNumString, nums.length);

  var styleString = hahPrefs.get(HAH_PREFS.HINT_STYLE);
  var styleStringFocus = hahPrefs.get(HAH_PREFS.HINT_STYLE_FOCUS);

  var xTemp, yTemp;

  if (oldLinkNumString.length > 0) {
    hintElem = hintElemArray[oldHintId];
    if (hahElemArray[oldHintId].tagName == 'FRAME' || hahElemArray[oldHintId].tagName == 'IFRAME') {
      hintElem.style.backgroundColor = 'ButtonFace';
    } else {
      xTemp = hintElem.style.left;
      yTemp = hintElem.style.top;
      hintElem.setAttribute('style', styleString);
      hintElem.style.position = 'absolute';
      hintElem.style.left = xTemp;
      hintElem.style.top = yTemp;
    }
  }
  if (linkNumString.length > 0) {
    hintElem = hintElemArray[hintId]; // TODO: has no properties??
    if (hahElemArray[hintId].tagName == 'FRAME' || hahElemArray[hintId].tagName == 'IFRAME') {
      hintElem.style.backgroundColor = 'ButtonShadow';
    } else {
      xTemp = hintElem.style.left;
      yTemp = hintElem.style.top;
      hintElem.setAttribute('style', styleStringFocus);
      hintElem.style.position = 'absolute';
      hintElem.style.left = xTemp; //(parseInt(xTemp) - (parseInt(widthTemp2) - parseInt(widthTemp)) / 2) + "px";
      hintElem.style.top = yTemp; //(parseInt(yTemp) - (parseInt(heightTemp2) - parseInt(heightTemp)) / 2) + "px";
    }
  }
}

function hahRegenHints()
{
  areCoordsValid = false;
}


////////////////////////////////////////////////////////////////////////////////
// basic functionality
////////////////////////////////////////////////////////////////////////////////

/**
 * Enables the HaH-mode by showing the hints and prepare for input the
 * hint numbers
 *
 * @author Pekka Sillanpaa
 * @param event that caused the mode to change
 * @return -1 if already enabled
 */

function hahEnableHahMode(event)
{
  if (isHahModeEnabled) 
    return -1;

  if (oWindowDocument != window._content.document) {
    areCoordsValid = false;
    getHintableElems();
  }

  oWindowDocument = window._content.document;

  hahStatusIcon("wait", true); // why doesn't work? xul don't refresh?
  linkNumString = '';
  isHahModeEnabled = true;
  hahShowHints();

  hahFAYTState = gUseTypeAheadFind; // disable FAYT
  if (gUseTypeAheadFind) gUseTypeAheadFind = false;

  hahStatusIcon("active", true);
  return 0;
}

/**
 * Disables the HaH-mode hiding the hints and disabling the input mode
 *
 * @author Pekka Sillanpaa
 * @param event that caused the mode to change
 * @param action = true if something is to be clicked
 *                 false if cancel
 * @return -1 if already disabled
 */
function hahDisableHahMode(event, action)
{
  if (!isHahModeEnabled) 
    return -1;

  hahRemoveHints();
  hahStatusIcon("enabled", true);
  hahStatusDisplay("");
  isHahModeEnabled = false;
  
  var nums = hahPrefs.get(HAH_PREFS.NUMBER_KEYS);
  var doc = window.document;
  
  var hintId = parseInt(linkNumString, nums.length);
  linkNumString = '';

  gUseTypeAheadFind = hahFAYTState; // re-enable FAYT if it was enabled

  if (hintId >= 0 && hintId < hahElemArray.length && action) {
    var evt = doc.createEvent('MouseEvents');
    var x = 0;
    var y = 0;

    var hahElemTagName = hahElemArray[hintId].tagName;

    if (hahElemTagName == 'FRAME' || hahElemTagName == 'IFRAME')
      return 0;

    // for imagemap
    if (hahElemTagName == 'AREA') {
      var coords = hahElemArray[hintId].getAttribute("coords").split(",");
      x = Number(coords[0]);
      y = Number(coords[1]);
    }
    evt.initMouseEvent('click', true, true, doc.defaultView, 1, x, y, 0, 0, event.ctrlKey, event.altKey, event.shiftKey, event.metaKey, 0, null);
    
    hahElemArray[hintId].dispatchEvent(evt);
    //    }

    // we don't want to remain the focus for radiobuttons and checkboxes
    if (hahElemArray[hintId].tagName == 'INPUT') {
      var type = hahElemArray[hintId].getAttribute("type");
      if (type == "radio" || type == "checkbox")
	hahBlurFocused();
    }
  } 
  return 0;
}

function hahSetMouseOverLink()
{
  if (linkNumString.length == 0) return;
  
  var doc = window.document;
  var nums = hahPrefs.get(HAH_PREFS.NUMBER_KEYS);
  
  var hintId = parseInt(linkNumString, nums.length);
  
  //  elem = linksArray[hintId];
  elem = hahElemArray[hintId];
  
  if (elem.tagName == 'FRAME' || elem.tagName == 'IFRAME') {
    elem.contentWindow.focus();
    return;
  } else {
    elem.focus();
  }
  
  if (hintId >= 0 && hintId < hintElemArray.length) {
    var evt = doc.createEvent('MouseEvents');
    var x = 0;
    var y = 0;
    // for imagemap
    if (elem.tagName == 'AREA') {
      var coords = elem.getAttribute("coords").split(",");
      x = Number(coords[0]);
      y = Number(coords[1]);
    }

    evt.initMouseEvent('mouseover', true, true, doc.defaultView, 1, x, y, 0, 0, 0, 0, 0, 0, 0, null);
    elem.dispatchEvent(evt);
  }
}

// not necessarily needed anymore
function generateKeypress(keyCode, charCode, shiftKey)
{
  var doc = getFirstDoc();
  var evt = doc.createEvent('KeyEvents');
  evt.initKeyEvent('keypress', true, true, window, false, false, shiftKey, false, keyCode, charCode, doc.defaultView);
  var elt = document.commandDispatcher.focusedElement;
  if (elt != null) elt.dispatchEvent(evt);
  else doc.body.dispatchEvent(evt);
}

function charcodeToNumber(charCode)
{
  var nums = hahPrefs.get(HAH_PREFS.NUMBER_KEYS).toLowerCase();
  var chr = String.fromCharCode(charCode).toLowerCase();
  var i = nums.indexOf(chr);
  if (i == -1) return null;
  return i.toString(nums.length);
}

function isFormElemFocused()
{
  if (hahPrefs.get(HAH_PREFS.KEYINPUT_IN_FORMS)) return false;

  var elt = document.commandDispatcher.focusedElement;
  if (elt == null) return false;

  var tagName = elt.localName.toUpperCase();

  //hahLogMessage(tagName);

  if (tagName == "INPUT" ||
      tagName == "TEXTAREA" ||
      tagName == "SELECT" ||
      tagName == "BUTTON" ||
      tagName == "ISINDEX")
    return true;
  
  return false;
}

function hahUpdateChanges(linkNumString, oldLinkNumString)
{
  //  hahStatusDisplay(linkNumString);
  hahChangeHintFocus(linkNumString, oldLinkNumString);
  hahSetMouseOverLink();
}

function hahTypeNumber(number)
{
  if (!isHahModeEnabled) return;

  var oldLinkNumString = null; 
  var nums = hahPrefs.get(HAH_PREFS.NUMBER_KEYS);
  
  if (number != null) {
    oldLinkNumString = linkNumString;
    linkNumString += '' + number;
    var hintId = parseInt(linkNumString, nums.length);
    if (!(hintId >= 0 && hintId < hintElemArray.length))
      linkNumString = oldLinkNumString;
  }
 
  hahUpdateChanges(linkNumString, oldLinkNumString);
}

function hahTypeBackspace() 
{
  var oldLinkNumString = null;
  if (linkNumString.length > 0) {
    oldLinkNumString = linkNumString;
    linkNumString = linkNumString.substr(0, linkNumString.length - 1);
  }
  hahUpdateChanges(linkNumString, oldLinkNumString);
}

////////////////////////////////////////////////////////////////////////////////
// event handlers
////////////////////////////////////////////////////////////////////////////////

function hahOnKeydown(event)
{

  if (!document.getElementById("FindToolbar").hidden) return; // disable HaH if find toolbar is open

  if (event.keyCode == 27 && !isHahModeEnabled) {
    hahBlurFocused();
    return;
  }

  if (isFormElemFocused() && !event.altKey) return;
    
  if (hahIsKey(event, hahPrefs.get(HAH_PREFS.START_KEY))) {
    event.preventDefault();
    event.stopPropagation();

    if (!isHahModeEnabled) {
      hahEnableHahMode(event);
    }
    else {
      hahDisableHahMode(event, false);
    }
  }    
  else if (hahIsKey(event, hahPrefs.get(HAH_PREFS.MAGIC_KEY))) {
    hahEnableHahMode(event);
    isMagicKeyPressed = true;
  }
}

function hahOnKeyup(event)
{
  if (!isHahModeEnabled && isFormElemFocused()) return;

  var generate = false;
  if (hahIsKey(event, hahPrefs.get(HAH_PREFS.MAGIC_KEY))) {
    if (linkNumString.length == 0) generate = true;

    hahDisableHahMode(event, true);
    event.preventDefault();
    event.stopPropagation();
    // if nothing happened to the number string, most propably user wanted press the MAGIC_KEY in the focused element
    if (hahPrefs.get(HAH_PREFS.GENERATE_MAGIC) && generate) generateKeypress(0, hahPrefs.get(HAH_PREFS.MAGIC_KEY) >> 4, event.shiftKey);
    isMagicKeyPressed = false;
  }
}

function hahOnKeypress(event)
{
  // if the window is zoomed
  if ((event.charCode == 45 || event.charCode == 43) && event.ctrlKey) {
    var p = false;
    hahDisableHahMode();
    hahResize();
  }

  if ((isFormElemFocused() && linkNumString == 0 && !event.altKey && event.keyCode != BACKSPACE_KEY) || !isHahModeEnabled) return;

  var num = charcodeToNumber(event.charCode);

  //if (event.keyCode ==       

  if (num != null) {
    event.preventDefault();
    event.stopPropagation(); 
    hahTypeNumber(num);
  }
  else if (event.keyCode == BACKSPACE_KEY && linkNumString.length > 0) {
    event.preventDefault();
    event.stopPropagation(); 
    hahTypeBackspace();    
    if (linkNumString == 0)
      hahDisableHahMode(event, false);
  } 
  else if (event.keyCode == STOP_KEY) {
    event.preventDefault();
    event.stopPropagation(); 
    hahDisableHahMode(event, false);
  }
  else if (isMagicKeyPressed) {
    event.preventDefault();
    event.stopPropagation(); 
  }
}

function hahIsKey(event, combinedKeyCode)
{
  var keyModifiers = combinedKeyCode & 0xF;
  var eventModifiers = event.altKey * Event.ALT_MASK |
    event.ctrlKey * Event.CONTROL_MASK |
    event.shiftKey * Event.SHIFT_MASK |
    event.metaKey * Event.META_MASK;
  // we don't care about modifiers if it's a keyup event
  if (event.type.toLowerCase() == 'keyup') keyModifiers = eventModifiers;
  return ((event.keyCode == combinedKeyCode >> 4) &&
          (keyModifiers == eventModifiers));
  
}

////////////////////////////////////////////////////////////////////////////////
// statusbar manipulating functions
////////////////////////////////////////////////////////////////////////////////

function hahStatusDisplay(string)
{
  var hahStatus = document.getElementById('hah-status');
  var status;
  if (string.length > 0) {
    hahStatus.setAttribute("status", "active");
    hahStatus.hidden = false;
  } else {
    hahStatus.setAttribute("status", "inactive");
    hahStatus.hidden = true;
  }
  hahStatus.label = string;
}

function hahStatusIcon(status, enabled)
{
  var hahStatus = document.getElementById('hah-status-icon');
  if (!enabled || !hahPrefs.get(HAH_PREFS.STATUSICON_ENABLED)) {
    hahStatus.hidden = true;
  } else {
    hahStatus.hidden = false;
    hahStatus.setAttribute("status", status);
  }
}

function windowStatusDisplay(string)
{
  window.status = string;
}

window.document.addEventListener("DOMContentLoaded", hahLoad, null);

//window.addEventListener("load", hahLoad, true);
window.document.addEventListener("resize", hahResize, true);


//is there any method to do this?
//would be nice with js popup menus etc.
//window.document.addEventListener("mutation", hahMutate, true);

// temporal solution
//window.document.addEventListener("click", hahResize, true);
window.document.addEventListener("change", hahResize, true);

//window.document.addEventListener("zoom", hahResize, true);
//window.addEventListener("blur", hahDisableHahMode, true);
//window.onblur = hahDisableHahMode;

hahLogMessage("Initialized.");

